# coding: utf8
from itertools import chain
import logging
import struct
import socket
from time import sleep
from functools import partial
from kafkaka.client import KafkaClient
from kafkaka.conn import Connection
from kafkaka.define import DEFAULT_POOL_SIZE, KafkaError, ConnectionError
from tornado.iostream import IOStream

logging.basicConfig(format='%(asctime)s:%(levelname)s:%(message)s', level=logging.DEBUG)
log = logging.getLogger("kafka")


class Gen():

    def __init__(self):
        pass

    @staticmethod
    def task(func, *args, **kwargs):
        # pack the function to be executed later
        return partial(func, *args, **kwargs)

    @staticmethod
    def callback(g, response=None):
        try:
            f = g.send(response)
            if hasattr(f, '__call__'):
                f(partial(Gen.callback, g))  # for multiple yield
        except StopIteration:
            pass

    @staticmethod
    def async(func):
        def wrapper(*args, **kwargs):
            g = func(*args, **kwargs)
            try:
                f = g.send(None)
            except StopIteration:
                pass
            else:
                if hasattr(f, '__call__'):
                    f(partial(Gen.callback, g))  # pack the generater in callback for later use
        return wrapper

gen = Gen()


class Connection(Connection):

    def __init__(self, pool=None, *args, **kwargs):
        super(Connection, self).__init__(*args, **kwargs)
        self._pool = pool
        self._stream = None
        self._callbacks = []
        self._ready = False

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._pool.release(self)

    def _add_callback(self, func):
        self._callbacks.append(func)

    def _do_callbacks(self):
        self._ready = True
        while 1:
            try:
                func = self._callbacks.pop()
                func()
            except IndexError:
                # all done
                break
            except:
                # other error
                continue

    def connect(self):
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM, 0)
        self._sock = IOStream(s)  # tornado iostream
        self._sock.connect((self._host, self._port), self._do_callbacks)

    def send(self, payload, correlation_id=-1, callback=None):
        """
        :param payload: an encoded kafka packet
        :param correlation_id: for now, just for debug logging
        :return:
        """
        if not self._ready:
            def _callback(*args, **kwargs):
                self.send(payload, correlation_id, callback)
            self._add_callback(_callback)
            return
        log.debug("About to send %d bytes to Kafka, request %d" % (len(payload), correlation_id))
        if payload:
            _bytes = struct.pack('>i%ds' % len(payload), len(payload), payload)
        else:
            _bytes = struct.pack('>i', -1)
        try:
            self._sock.write(_bytes, callback)  # simply using sendall
        except:
            self.close()
            callback(None)
            self._log_and_raise('Unable to send payload to Kafka')

    def _recv(self, size, callback):
        try:
            self._sock.read_bytes(min(size, 4096), callback)
        except:
            self.close()
            callback(None)  # if error, set None
            self._log_and_raise('Unable to receive data from Kafka')

    def recv(self, correlation_id=-1, callback=None):
        """

        :param correlation_id: for now, just for debug logging
        :return: kafka response packet
        """
        log.debug("Reading response %d from Kafka" % correlation_id)
        if not self._ready:
            def _callback():
                self.recv(correlation_id, callback)
            self._add_callback(_callback)
            return
        def get_size(resp):
            if resp == None:
                callback(None)
            size, = struct.unpack('>i', resp)
            self._recv(size, callback)
        self._recv(4, get_size)  # read the response length

    def close(self):
        self._callbacks = []
        log.debug("Closing socket connection" + self._log_tail)
        if self._sock:
            self._sock.close()
            self._sock = None
        else:
            log.debug("Socket connection not exists" + self._log_tail)

    def closed(self):
        return self._sock.closed()


class ConnectionPool(object):
    def __init__(self, connection_class=Connection, **connection_kwargs):
        self.connection_class = connection_class
        self.connection_kwargs = connection_kwargs
        self.reset()

    def __repr__(self):
        return "%s<%s%s>" % (
            type(self).__name__,
            self.connection_class.__name__,
            str(self.connection_kwargs)
        )

    def __len__(self):
        return len(self._available_connections) + len(self._in_use_connections)

    def reset(self):
        self._created_connections = 0
        self._available_connections = []
        self._in_use_connections = set()

    def get_connection(self):
        """
        Get a connection from the pool
        :return: Connection object
        """
        try:
            c = self._available_connections.pop()
        except IndexError:
            c = self.connection_class(pool=self, **self.connection_kwargs)
        if c.closed():
            self.reset()
            return self.get_connection()
        self._in_use_connections.add(c)
        return c

    def release(self, connection):
        """
        Releases the connection back to the pool
        :param connection:
        :return:
        """
        if connection in self._in_use_connections:
            self._in_use_connections.remove(connection)
            self._available_connections.append(connection)


    def disconnect(self):
        """
        Disconnects all connections in the pool
        :return:
        """
        all_conns = chain(self._available_connections,
                          self._in_use_connections)
        for connection in all_conns:
            connection.disconnect()


class KafkaClient(KafkaClient):

    def __init__(self, *args, **kwargs):
        self._pools = {}
        super(KafkaClient, self).__init__(*args, **kwargs)

    def _get_conn(self, host, port):
        """
        Get or create a connection using Pool
        :param host: host name
        :param port: port number
        :return: Connection
        """
        key = (host, port)
        if key not in self._pools:
            self._pools[key] = ConnectionPool(host=host, port=port)
        return self._pools[key].get_connection()

    @gen.async
    def boot_metadata(self, expected_topics, callback=None):
        """
        boot metadata from kafka server
        :param expected_topics: The topics to produce metadata for. If empty the request will yield metadata for all topics.
        :return:
        """
        correlation_id, _topics, request_bytes = self._pack_boot_metadata(expected_topics)
        for (host, port) in self.hosts*3:  # trick for auto-create topics
            try:
                with self._get_conn(host, port) as conn:
                    yield gen.task(conn.send, request_bytes, correlation_id)
                    resp_bytes = yield gen.task(conn.recv, correlation_id)
                self._unpack_boot_metadata(resp_bytes, expected_topics, callback)
                return  # break the loop
            except KafkaError as e:
                log.warning("Kafka metadata via request [%r] from server %s:%i is not available, "
                            "trying next server: %s" % (correlation_id, host, port, e))
                sleep(1)  # trick for auto-create topics
                continue
            except Exception as e:
                log.warning("Could not send request [%r] to server %s:%i, "
                            "trying next server: %s" % (correlation_id, host, port, e))
                continue

        raise KafkaError("All servers failed to process request")

    @gen.async
    def send_message(self, topic_name, *msg):
        if not self._ready:
            def _callback():
                self.send_message(topic_name, *msg)
            self._add_callback(_callback)
            return
        host, port, request_bytes, correlation_id = self._pack_send_message(topic_name, *msg)
        for i in xrange(self._retry_times):
            with self._get_conn(host, port) as conn:
                try:
                    yield gen.task(conn.send, request_bytes, correlation_id)
                except ConnectionError as e:
                    log.warning("Could not send request [%r] to server %s:%i, try again, %s" % (correlation_id, host, port, e))
                    if i == self._retry_times - 1:
                        log.error("Could not send request [%r] to server %s:%i, %s" % (correlation_id, host, port, e))
                    continue  # try more
                try:
                    resp_bytes = yield gen.task(conn.recv, correlation_id)
                    self._unpack_send_message(resp_bytes)
                except ConnectionError as e:
                    log.error('Could not get response [%r] from server %s:%i, %s' % (correlation_id, host, port, e))
                except Exception as e:
                    log.error("Bad response [%r] from server %s:%i, %s" % (correlation_id, host, port, e))
            break

if __name__ == "__main__":
    pass


# coding: utf8

from kafkaka.bstruct import (
    Struct,
    ShortField, IntField, CharField, UnsignedCharField,
    LongLongField, UnsignedIntField, Crc32Field,
)


class KafkaProtocol(object):
    PRODUCE_KEY = 0
    FETCH_KEY = 1
    OFFSET_KEY = 2
    METADATA_KEY = 3
    OFFSET_COMMIT_KEY = 8
    OFFSET_FETCH_KEY = 9
    CLIENT_ID = 'kafkaka'


class HeadStruct(Struct):
    api_key = ShortField()
    api_version = ShortField(default=0)
    correlation_id = IntField()
    client_id_length = ShortField(default=len(KafkaProtocol.CLIENT_ID))
    client_id = CharField(default=KafkaProtocol.CLIENT_ID, length='client_id_length')


class MetaTopicStruct(Struct):
    topic_name_length = ShortField()
    topic_name = CharField(length='topic_name_length')


class MetaStruct(Struct):
    head = HeadStruct(api_key=KafkaProtocol.METADATA_KEY)
    topic_length = IntField(default=0)
    topics = MetaTopicStruct(repeat='topic_length')


class BrokerStruct(Struct):
    node_id = IntField()
    host_char_len = ShortField()
    host = CharField(length='host_char_len')
    port = IntField()


class PartitionStruct(Struct):
    error = ShortField()
    partition = IntField()
    leader = IntField()
    replicas_number = IntField()
    replicas = IntField(repeat='replicas_number')
    isr_number = IntField()
    isr = IntField(repeat='isr_number')


class TopicStruct(Struct):
    error = ShortField()
    topic_name_length = ShortField()
    topic_name = CharField(length='topic_name_length')
    partitions_number = IntField()
    partitions = PartitionStruct(repeat='partitions_number')


class MetaResponseStruct(Struct):
    correlation_id = IntField()
    brokers_number = IntField()
    brokers = BrokerStruct(repeat='brokers_number')
    topics_number = IntField()
    topics = TopicStruct(repeat='topics_number')


class MessageStruct(Struct):
    magic = UnsignedCharField(default=0)
    attributes = UnsignedCharField(default=0)
    key = IntField(default=-1)
    value_length = IntField()
    value = CharField(length='value_length')


class Crc32MessageStruct(Struct):
    crc = Crc32Field(source='message')
    message = MessageStruct()


class Crc32MessageStructWithLength(Struct):
    unknown = LongLongField(default=0)
    message_length = IntField(source='message')
    message = Crc32MessageStruct()


class MessageSetStruct(Struct):
    partition = IntField()
    message_set_length = IntField(source='message_set')
    message_set = Crc32MessageStructWithLength(repeat='message_set_length')


class ProducePayloadStruct(Struct):
    topic_name_length = ShortField()
    topic_name = CharField(length='topic_name_length')
    topic_payloads_number = IntField()
    topic_payloads = MessageSetStruct(repeat='topic_payloads_number')


class ProduceStruct(Struct):
    head = HeadStruct(KafkaProtocol.PRODUCE_KEY, 0)
    acks = ShortField(default=1)
    timeout = IntField(default=1000)
    payloads_number = IntField()
    payloads = ProducePayloadStruct(repeat='payloads_number')


class ResponsePartitionStruct(Struct):
    partition = IntField()
    error = ShortField()
    offset = LongLongField()


class ResponseTopicStruct(Struct):
    topic_name_length = ShortField()
    topic_name = CharField(length='topic_name_length')
    partitions_number = IntField()
    partitions = ResponsePartitionStruct(repeat='partitions_number')


class ProduceResponseStruct(Struct):
    correlation_id = IntField()
    topics_number = IntField()
    topics = ResponseTopicStruct(repeat='topics_number')


if __name__ == "__main__":

    s = HeadStruct(
        KafkaProtocol.METADATA_KEY,
        0,
        111,
    )
    print s.pack2bin()
    s2 = HeadStruct(
        KafkaProtocol.METADATA_KEY,
        0,
        112,
        client_id='ok'
    )
    print s2.pack2bin()
    m1 = MetaStruct(
        head=dict(correlation_id=111)
    )
    print m1.pack2bin()

    m2 = MetaStruct(
        head=dict(correlation_id=222)
    )
    print m2.pack2bin()
    binary = m1.pack2bin()
    m3 = MetaStruct()
    print 'unpack'
    m3.unpack(binary)
    print m3.pack2bin()
    print m3.dump2nametuple()

    msg = MessageStruct(
        value='test'
    )
    print (msg.dump2nametuple(), msg.pack2bin())
    crc_msg = Crc32MessageStruct(
        message=dict(
            value='test'
        )
    )
    assert crc_msg.pack2bin() == 'Y*G\x87\x00\x00\xff\xff\xff\xff\x00\x00\x00\x04test'


    crc_msg_with_length = Crc32MessageStructWithLength(
        message=dict(
            message=dict(
                value='test'
            )
        )
    )
    assert crc_msg_with_length.pack2bin() == '\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x12Y*G\x87\x00\x00\xff\xff\xff\xff\x00\x00\x00\x04test'

    msg_set = MessageSetStruct(
        partition=1,
        message_set=[
            dict(
                message=dict(
                    message=dict(
                        value='test'
                    )
                )
            ),
        ]
    )
    msg_set.pack2bin()

    produce_msg = ProduceStruct(
        head = dict(correlation_id=1),
        payloads=[
            dict(
                topic_name='im-msg',
                topic_payloads=[
                    dict(
                        partition=0,
                        message_set=[
                            dict(
                                message=dict(
                                    message=dict(
                                        value='test'
                                    )
                                )
                            ),
                        ]
                    ),
                ],
            ),
        ]
    )
    assert produce_msg.pack2bin() == '\x00\x00\x00\x00\x00\x00\x00\x01\x00\x07kafkaka\x00\x01\x00\x00\x03\xe8\x00\x00\x00\x01\x00\x06im-msg\x00\x00\x00\x01\x00\x00\x00\x00\x00\x00\x00\x1e\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x12Y*G\x87\x00\x00\xff\xff\xff\xff\x00\x00\x00\x04test'
    out = ProduceStruct()
    out.unpack(produce_msg.pack2bin())
    print out.dump2nametuple()
    assert out.dump2nametuple().payloads[0].topic_name == 'im-msg'


# coding: utf8

#
# config
#

DEFAULT_SOCKET_TIMEOUT_SECONDS = 30
DEFAULT_KAFKA_PORT = 9092
DEFAULT_RETRY_TIMES = 3
DEFAULT_POOL_SIZE = None

#
# custom errors
# see: https://cwiki.apache.org/confluence/display/KAFKA/A+Guide+To+The+Kafka+Protocol#AGuideToTheKafkaProtocol-ErrorCodes


class KafkaError(Exception):
    pass


class ConnectionError(KafkaError):
    """
    socket connection error
    """


class NoError(KafkaError):
    """
    No error--it worked!
    """
    code = 0


class Unknown(KafkaError):
    """
    An unexpected server error
    """
    code = -1


class OffsetOutOfRange(KafkaError):
    """
    The requested offset is outside the range of offsets maintained by the server for the given topic/partition.
    """
    code = 1


class InvalidMessage(KafkaError):
    """
    This indicates that a message contents does not match its CRC
    """
    code = 2


class UnknownTopicOrPartition(KafkaError):
    """
    This request is for a topic or partition that does not exist on this broker.
    """
    code = 3


class InvalidMessageSize(KafkaError):
    """
    The message has a negative size
    """
    code = 4


class LeaderNotAvailable(KafkaError):
    """
    This error is thrown if we are in the middle of a leadership election and
    there is currently no leader for this partition and hence it is unavailable for writes.
    """
    code = 5


class NotLeaderForPartition(KafkaError):
    """
    This error is thrown if the client attempts to send messages to a replica that
    is not the leader for some partition. It indicates that the clients metadata is out of date.
    """
    code = 6


class RequestTimedOut(KafkaError):
    """
    This error is thrown if the request exceeds the user-specified time limit in the request.
    """
    code = 7


class BrokerNotAvailable(KafkaError):
    """
    This is not a client facing error and is used mostly by tools when a broker is not alive.
    """
    code = 8


class ReplicaNotAvailable(KafkaError):
    """
    If replica is expected on a broker, but is not.
    """
    code = 9


class MessageSizeTooLarge(KafkaError):
    """
    The server has a configurable maximum message size to avoid unbounded memory allocation.
    This error is thrown if the client attempt to produce a message larger than this maximum.
    """
    code = 10


class StaleControllerEpochCode(KafkaError):
    """
    Internal error code for broker-to-broker communication.
    """
    code = 11


class OffsetMetadataTooLargeCode(KafkaError):
    """
    If you specify a string larger than configured maximum for offset metadata
    """
    code = 12


class OffsetsLoadInProgressCode(KafkaError):
    """
    The broker returns this error code for an offset fetch request
    if it is still loading offsets (after a leader change for that offsets topic partition).
    """
    code = 14


class ConsumerCoordinatorNotAvailableCode(KafkaError):
    """
    The broker returns this error code for consumer metadata requests
    or offset commit requests if the offsets topic has not yet been created.
    """
    code = 15


class NotCoordinatorForConsumerCode(KafkaError):
    """
    The broker returns this error code if it receives an offset fetch
    or commit request for a consumer group that it is not a coordinator for.
    """
    code = 16


error_codes = {
    -1: Unknown,
    0: NoError,
    1: OffsetOutOfRange,
    2: InvalidMessage,
    3: UnknownTopicOrPartition,
    4: InvalidMessageSize,
    5: LeaderNotAvailable,
    6: NotLeaderForPartition,
    7: RequestTimedOut,
    8: BrokerNotAvailable,
    9: ReplicaNotAvailable,
    10: MessageSizeTooLarge,
    11: StaleControllerEpochCode,
    12: OffsetMetadataTooLargeCode,
    14: OffsetsLoadInProgressCode,
    15: ConsumerCoordinatorNotAvailableCode,
    16: NotCoordinatorForConsumerCode,
}


def check_and_raise_error(response):
    error = error_codes.get(response.error, Unknown)
    if error is not NoError:
        raise error(response)


# coding: utf8
from gevent import monkey
from itertools import chain
import logging
from kafkaka.client import KafkaClient
from kafkaka.conn import Connection
from kafkaka.define import DEFAULT_POOL_SIZE
from gevent import spawn
from gevent.queue import Queue
import socket
monkey.patch_socket()
log = logging.getLogger("kafka")


class Connection(Connection):

    def __init__(self, pool=None, *args, **kwargs):
        super(Connection, self).__init__(*args, **kwargs)
        self._pool = pool

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._pool.release(self)

    def close(self):
        super(Connection, self).close()
        self._pool = None


class ConnectionPool(object):
    def __init__(self, connection_class=Connection, pool_size=DEFAULT_POOL_SIZE, **connection_kwargs):
        self.connection_class = connection_class
        self.connection_kwargs = connection_kwargs
        self.pool_size = pool_size  # the number of max parallel connections
        self.reset()

    def __repr__(self):
        return "%s<%s%s>" % (
            type(self).__name__,
            self.connection_class.__name__,
            str(self.connection_kwargs)
        )

    def __len__(self):
        return len(self._available_connections) + len(self._in_use_connections)

    def reset(self):
        self._created_connections = 0
        self._available_connections = []
        self._in_use_connections = set()
        if self.pool_size is not None:
            self.q = Queue(maxsize=self.pool_size)
            for i in xrange(self.pool_size):
                self.q.put_nowait(i)
        else:
            self.q = None

    def get_connection(self):
        """
        Get a connection from the pool
        :return: Connection object
        """
        if self.q:
            self.q.get()
        try:
            c = self._available_connections.pop()
        except IndexError:
            c = self.connection_class(pool=self, **self.connection_kwargs)
        if c.closed():
            self.reset()
            return self.get_connection()
        self._in_use_connections.add(c)
        return c

    def release(self, connection):
        """
        Releases the connection back to the pool
        :param connection:
        :return:
        """
        if self.q:
            self.q.put_nowait(1)
        if connection in self._in_use_connections:
            self._in_use_connections.remove(connection)
            self._available_connections.append(connection)


    def disconnect(self):
        """
        Disconnects all connections in the pool
        :return:
        """
        all_conns = chain(self._available_connections,
                          self._in_use_connections)
        for connection in all_conns:
            connection.disconnect()


class KafkaClient(KafkaClient):

    def __init__(self, *args, **kwargs):
        self._pools = {}
        self._pool_size = kwargs.pop('pool_size', DEFAULT_POOL_SIZE)
        super(KafkaClient, self).__init__(*args, **kwargs)

    def _get_conn(self, host, port):
        """
        Get or create a connection using Pool
        :param host: host name
        :param port: port number
        :return: Connection
        """
        key = (host, port)
        if key not in self._pools:
            self._pools[key] = ConnectionPool(host=host, port=port, pool_size=self._pool_size)
        return self._pools[key].get_connection()

    def send_message(self, *args, **kwargs):
        return spawn(super(KafkaClient, self).send_message, *args, **kwargs)


if __name__ == "__main__":
    pass


# coding: utf8
from kafkaka.gevent_patch import KafkaClient
from gevent import joinall

import time

if __name__ == "__main__":
    c = KafkaClient("t-storm1:9092",
                    topic_names=['im-msg'],
                    pool_size=10  # the number of max parallel connections.
    )
    start = time.time()
    all = []
    print ''
    for i in xrange(50):
        all.append(c.send_message('im-msg', u'你好', str(time.time()), i))
        all.append(c.send_message('im-msg', 'hi', str(time.time()), i))
    print 'this will not block'
    for i in xrange(50):
        all.append(c.send_message('im-msg', u'你好', str(time.time()), i))
        all.append(c.send_message('im-msg', 'hi', str(time.time()), i))
    joinall(all)
    print 'but this will block'
    print time.time() - start
